#include <ruby.h>
#include <string.h>

#ifdef ON_WINDOWS
#include <windows.h>
#include <shlwapi.h>
#include <tchar.h>
#ifndef UNICODE
#define UNICODE
#endif
#endif

#ifndef MAX_PATH
#define MAX_PATH 260
#endif

/* Sometimes it's easier to deal with a string, sometimes a char */
#ifdef ON_WINDOWS
#define SEPS "\\"
#define SEPC '\\'
#else
#define SEPS "/"
#define SEPC '/'
#endif

static VALUE path_root(VALUE);
static VALUE path_to_a(VALUE);
static VALUE path_clean(VALUE);
static VALUE path_clean_bang(VALUE);

#ifdef ON_WINDOWS
static VALUE path_is_unc(VALUE);
#endif

VALUE cPathname, cPathnameError;
VALUE cFileUtils;

/* Helper functions */
static VALUE yield_block(VALUE v_arg, VALUE v_block){
   return rb_funcall2(v_block, rb_intern("call"), 1, &v_arg); 
}

static VALUE call_io_foreach(VALUE v_args){
   return rb_funcall2(rb_cIO, rb_intern("foreach"), 2, (VALUE*)v_args);  
}

static VALUE call_dir_open(VALUE v_args){
   return rb_funcall2(rb_cDir, rb_intern("open"), 1, (VALUE*)v_args);  
}

static VALUE call_file_open(VALUE v_args){
   return rb_funcall2(rb_cFile, rb_intern("open"), 2, (VALUE*)v_args);  
}

/* 
 * Pathname.new(path)
 *
 * Creates and and returns a new Pathname object.  On Windows, all forward
 * slashes are replaced with backslashes.  In addition, file urls are
 * converted to normalized paths on Windows.
 */
static VALUE path_initialize(VALUE self, VALUE v_string){      
#ifdef ON_WINDOWS
   TCHAR* path = TEXT(StringValuePtr(v_string));
   
   if(PathIsURL(path)){    
      HRESULT hr;
      TCHAR buf[MAX_PATH];
      DWORD dwLength = sizeof(buf);
      
      hr = PathCreateFromUrl(path, buf, &dwLength, 0);
      
      if(hr != S_OK)
         rb_raise(cPathnameError, "invalid file url: %s", path);
         
      v_string = rb_str_new2(buf);  
   }
      
   /* Convert forward slashes to backslashes */
   v_string = rb_funcall(v_string, rb_intern("tr"), 2,
      rb_str_new2("/"), rb_str_new2(SEPS)
   );
#else
   char* path = StringValuePtr(v_string);
   if(strncmp(path, "file://", 7) == 0){
      VALUE v_temp, v_uri, cURI;

      rb_require("uri");

      cURI = rb_const_get(rb_cObject, rb_intern("URI"));
      v_uri = rb_funcall(cURI, rb_intern("parse"), 1, v_string);
      v_temp = rb_funcall(v_uri, rb_intern("path"), 0); 
      v_string = rb_funcall(cURI, rb_intern("decode"), 1, v_temp);
   }
#endif

   return rb_call_super(1, &v_string);
}

/*
 * Pathname#to_a
 *
 * Splits a pathname into pieces based on the path separator.  For example,
 * "/foo/bar/baz" would return a three element array of ['foo', 'bar', 'baz']
 * 
 * :no-doc:
 * I use rb_funcall() instead of rb_ary_delete() because the latter was
 * causing the deleted elements to be yielded in other methods where I use this
 * method as a helper.
 * 
 * I'd love to know whose bright idea it was to allow a block on Array#delete.
 */
static VALUE path_to_a(VALUE self){
   VALUE v_array = rb_str_split(self, SEPS);
   rb_funcall(v_array, rb_intern("delete"), 1, rb_str_new2(""));
   
   return v_array;
}

/*
 * Pathname#root
 *
 * Returns the root directory of the path, or '.' if there is no root
 * directory.
 * 
 * On Unix, this means the '/' character.  On Windows systems, this can
 * refer to the drive letter, or the server and share path if the path
 * is a UNC path.
 */
static VALUE path_root(VALUE self){
   VALUE v_root = rb_str_new2(".");

   if(RSTRING(self)->len > 0){
#ifdef ON_WINDOWS
      TCHAR szRoot[MAX_PATH];
   
      _tcsncpy(szRoot, TEXT(RSTRING(self)->ptr), RSTRING(self)->len + 1);

      if(PathStripToRoot(szRoot))
         v_root = rb_str_new2(szRoot);
#else
      if(RSTRING(self)->ptr[0] == SEPC)
         v_root = rb_str_new2(SEPS);
#endif
   }

   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, v_root);
}

/* 
 * Pathname#root?
 * 
 * Returns whether or not the path consists only of a root directory.
 */
static VALUE path_is_root(VALUE self){
   VALUE v_bool = Qfalse;
   
#if ON_WINDOWS
   if(PathIsRoot(TEXT(RSTRING(self)->ptr)))      
      v_bool = Qtrue;
#else
   VALUE v_root = path_root(self);

   if(!strcmp(StringValuePtr(self), StringValuePtr(v_root)))
      v_bool = Qtrue;
#endif

   return v_bool;
}

/*
 * Pathname#relative?
 *
 * Returns whether or not the path is a relative path.
 */
static VALUE path_is_relative(VALUE self){
   VALUE v_bool = Qfalse;
   
#ifdef ON_WINDOWS
   if(PathIsRelative(TEXT(RSTRING(self)->ptr)))
      v_bool = Qtrue;
#else
   VALUE v_root = path_root(self);
 
   if(!strcmp(StringValuePtr(v_root), "."))
      v_bool = Qtrue;
#endif

   return v_bool;
}

/*
 * Pathname#absolute?
 *
 * Returns whether or not the path is an absolute path.
 */
static VALUE path_is_absolute(VALUE self){
   VALUE v_bool = path_is_relative(self);

   /* Return the opposite of path_is_relative() */
   if(RTEST(v_bool))
      v_bool = Qfalse;
   else
      v_bool = Qtrue;

   return v_bool;
}

/*
 * Pathname#pstrip!
 *
 * Removes the trailing slash if present, modifying the receiver in place.
 */
static VALUE path_pstrip_bang(VALUE self){
#ifdef ON_WINDOWS
   TCHAR* path = TEXT(StringValuePtr(self));
      
   PathRemoveBackslash(path);
           
   rb_funcall(self, rb_intern("replace"), 1, rb_str_new2(path));
#else
   int last = RSTRING(self)->len - 1;

   if(RSTRING(self)->ptr[last] == SEPC)
      rb_funcall(self, rb_intern("chop!"), 0, 0);
      
#endif 
   return self;
}

/*
 * Pathname#pstrip
 *
 * Removes the trailing slash if present.  Does not modify the receiver.
 */
static VALUE path_pstrip(VALUE self){
   VALUE v_temp;
#ifdef ON_WINDOWS
   TCHAR path[MAX_PATH];
#endif

   v_temp = rb_str_dup(self);
   rb_str_modify(v_temp);

#ifdef ON_WINDOWS
   _tcsncpy(path, TEXT(StringValuePtr(v_temp)), RARRAY(v_temp)->len+1);
      
   PathRemoveBackslash(path);
           
   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, rb_str_new2(path));
#else
   int last = RSTRING(self)->len - 1;

   if(RSTRING(v_temp)->ptr[last] == SEPC)
      rb_funcall(v_temp, rb_intern("chop!"), 0, 0);
      
   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, v_temp);
#endif 
}

/*
 * Pathname#+
 *
 * Adds two Pathname objects together, or a Pathname and a String, returning
 * a new Pathname object.  It also automatically cleans the Pathname.
 *
 * Example:
 *    path1 = '/foo/bar'
 *    path2 = '../baz'
 *    path1 + path2 # '/foo/baz'
 *
 * Adding a root path to an existing path merely replaces the current
 * path.  Adding '.' to an existing path does nothing.
 */
static VALUE path_add(VALUE self, VALUE v_string){
   VALUE v_new_string, v_dup_self, v_dup_string;
#ifdef ON_WINDOWS
   TCHAR lpszDst[MAX_PATH];
   TCHAR path[MAX_PATH];
   TCHAR more[MAX_PATH];
#else
   VALUE v_arr;
#endif

   /* Don't modify left or right operand */
   v_dup_self   = rb_str_dup(self);
   v_dup_string = rb_str_dup(v_string);

   rb_str_modify(v_dup_self);
   rb_str_modify(v_dup_string);

#ifdef ON_WINDOWS
   _tcsncpy(path, TEXT(StringValuePtr(v_dup_self)), RARRAY(self)->len+1);
   _tcsncpy(more, TEXT(StringValuePtr(v_dup_string)), RARRAY(v_string)->len+1);
#endif

   /* Convert plain string to Pathname */
   if(TYPE(v_dup_string) != TYPE(self))
      v_dup_string = rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, v_dup_string);

   /* Return self if v_dup_string is '.' */
   if(!strcmp(StringValuePtr(v_dup_string), "."))
      return self;

   /* Return the string if self is '.' */
   if(!strcmp(StringValuePtr(self), "."))
      return v_dup_string;

#ifdef ON_WINDOWS 
   if(!PathAppend(path, more))
      rb_raise(cPathnameError, "unable to append string %s", more);
      
   if(!PathCanonicalize(lpszDst, path))
      rb_raise(cPathnameError, "PathCanonicalize() failed");

   v_new_string = rb_str_new2(lpszDst);
#else
   /* Return the string as-is if it's an absolute string */
   if(RTEST(path_is_absolute(v_string)))
      return v_string;

   v_arr = path_to_a(v_dup_self);
   rb_ary_concat(v_arr, path_to_a(v_string));

   v_new_string = rb_funcall(v_arr, rb_intern("join"), 1, rb_str_new2(SEPS));

   if(!RTEST(path_is_relative(v_dup_self)))
      v_new_string = rb_funcall(rb_str_new2(SEPS), rb_intern("+"), 1, v_new_string);
#endif

   v_new_string = path_clean(v_new_string);
   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, v_new_string);
}

/*
 * Pathname#clean
 *
 * Removes unnecessary '.' paths and ellides '..' paths appropriately.
 * Non-destructive.
 */
static VALUE path_clean(VALUE self){
   VALUE v_temp  = rb_str_dup(self);
#ifdef ON_WINDOWS
   TCHAR lpszSrc[MAX_PATH];
   TCHAR lpszDst[MAX_PATH];
#else
   int i;
   VALUE v_arr, v_element, v_root;
   VALUE v_final = rb_ary_new();
#endif

   rb_str_modify(v_temp);

#ifdef ON_WINDOWS
   _tcsncpy(lpszSrc, TEXT(StringValuePtr(v_temp)), RARRAY(v_temp)->len+1);
#endif
   
   /* Return self if the string is empty */
   if(RSTRING(v_temp)->len == 0)
      return v_temp;
   
#ifdef ON_WINDOWS  
   if(PathCanonicalize(lpszDst, lpszSrc))
      return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, rb_str_new2(lpszDst));
   else
      return self;
#else
   v_arr  = path_to_a(v_temp);
   v_root = path_root(v_temp);

   for(i = 0; i < RARRAY(v_arr)->len; i++){
      v_element = RARRAY(v_arr)->ptr[i];

      if(!rb_str_cmp(v_element, rb_str_new2("."))){
         continue;
      }

      rb_ary_push(v_final, v_element);

      if( (!rb_str_cmp(v_element, rb_str_new2(".."))) &&
          (rb_str_cmp(self, rb_str_new2("..")))
      )
      {
         rb_ary_pop(v_final);
         rb_ary_pop(v_final);
      }
   } 

   v_final = rb_ary_join(v_final, rb_str_new2(SEPS));

   if(rb_str_cmp(v_root, rb_str_new2("."))){
      v_final = rb_str_plus(v_root, v_final);
   }

   /* If the final string is empty, return '.' */
   if(RSTRING(v_final)->len == 0)
      v_final = rb_str_new2(".");

   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, v_final);
#endif
}

/*
 * Pathname#clean!
 *
 * Removes unnecessary '.' paths and ellides '..' paths appropriately.
 * Modifies the receiver in place.
 */
static VALUE path_clean_bang(VALUE self){
   rb_funcall(self, rb_intern("replace"), 1, path_clean(self));
   return self;
}

/*
 * Pathname#each
 *
 * Yields each component of the path name to a block.
 */
static VALUE path_each(VALUE self){
   VALUE v_array = path_to_a(self);
   int i;
   
   for(i = 0; i < RARRAY(v_array)->len; i++)
      rb_yield(RARRAY(v_array)->ptr[i]);

   return self;  
}

/*
 * Pathname#ascend
 *
 * Yields the path, minus one component on each iteration, as a new
 * Pathname object, ending with the root path.
 *
 * For example, if the path is "/foo/bar/baz", then "/foo/bar/baz" would
 * be yielded on the first iteration, "/foo/bar" on the second, "/foo" on
 * the third, and finally "/".
 */
static VALUE path_ascend(VALUE self){
   VALUE v_path, v_array, v_temp;
   int i, n;

   if(RTEST(path_is_root(self))){
      rb_yield(self);
      return Qnil;  
   }

   v_array = path_to_a(self);
   n = RARRAY(v_array)->len;

   while(n > 0){
      v_temp = rb_ary_new();
      for(i = 0; i < n; i++){
         rb_ary_push(v_temp, RARRAY(v_array)->ptr[i]);
      }

      v_path = rb_funcall(v_temp, rb_intern("join"), 1, rb_str_new2(SEPS));

      if(RTEST(path_is_absolute(self))){
#ifdef ON_WINDOWS
         if(RTEST(path_is_unc(self))){
            v_path = rb_str_concat(rb_str_new2("\\\\"), v_path);
         }
#else
         v_path = rb_str_concat(path_root(self), v_path);
#endif
      }

      rb_yield(rb_funcall(cPathname, rb_intern("new"), 1, v_path));

#ifdef ON_WINDOWS
      if( (RTEST(path_is_unc(self))) && (RTEST(path_is_root(v_path))) ){
         break;
      }
#endif
      n--;
   }

#ifndef ON_WINDOWS
   if(RTEST(path_is_absolute(self)))
      rb_yield(path_root(self));
#endif

   return self;

}

/* 
 * Pathname#descend
 *
 * Yields each component of the path, concatenating the next component on
 * each iteration, as a new Pathname object, starting with the root path.
 *
 * For example, if the path is "/foo/bar/baz", then "/" would be yielded
 * on the first iteration, "/foo" on the second, "/foo/bar on the third and
 * finally "/foo/bar/baz".
 */
static VALUE path_descend(VALUE self){
   VALUE v_path, v_array, v_element;
   int i;
   
   v_path = rb_str_new2("");
   
   if(RTEST(path_is_root(self))){
      rb_yield(self);
      return Qnil;  
   }

#ifdef ON_WINDOWS
   if(RTEST(path_is_unc(self))){
      rb_str_concat(v_path, path_root(self));
      rb_str_concat(v_path, rb_str_new2(SEPS));    
   }
#else
   if(RTEST(path_is_absolute(self)))
      v_path = path_root(self);
#endif

#ifdef ON_WINDOWS
   if(RTEST(path_is_unc(self))){
      if(RTEST(path_is_absolute(self))){
         rb_yield(path_root(self));
      }
   }
#else
   if(RTEST(path_is_absolute(self)))
      rb_yield(path_root(self));
#endif

   v_array = path_to_a(self);
   
   for(i = 0; i < RARRAY(v_array)->len; i++){
      v_element = RARRAY(v_array)->ptr[i];
#ifdef ON_WINDOWS
      if(RTEST(path_is_unc(self))){
         VALUE v_root = path_root(self);
         VALUE v_temp = rb_funcall(path_root(self), rb_intern("to_a"), 0, 0);
         VALUE v_bool = rb_funcall(v_temp, rb_intern("include?"), 1, v_element);
         
         if(RTEST(v_bool))
            continue;   
      }
#endif
      rb_str_concat(v_path, v_element);
      rb_str_concat(v_path, rb_str_new2(SEPS));
      rb_yield(rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1,
         rb_funcall(v_path, rb_intern("chop"), 0, 0))
      );
   }
   
   return self;  
}

/*
 * Pathname#<=>
 *
 * Pathnames can only be compared against other Pathnames.  Otherwise
 * nil is returned.
 */
static VALUE path_compare(VALUE self, VALUE v_string){
   if(TYPE(v_string) != TYPE(self))
      return Qnil;

   return rb_call_super(1, &v_string);
}

/*
 * Pathname#realpath
 * 
 * Returns a real (absolute) pathname of +self+ in the actual filesystem.
 *
 * Unlike most Pathname methods, this one assumes that the path actually
 * exists on your filesystem.  If it doesn't, an error is raised.
 *--
 * This method uses your system's realpath() function if it has one.
 */
static VALUE path_realpath(VALUE self){
#ifdef HAVE_REALPATH
   char buf[MAX_PATH];
   if(!realpath(RSTRING(self)->ptr, buf)){
      rb_sys_fail("realpath");
   }
   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, rb_str_new2(buf));
#else
   VALUE v_path, v_pwd;

   rb_funcall(rb_cFile, rb_intern("stat"), 1, self); /* Ensure path exists */

   if(RTEST(rb_funcall(rb_cFile, rb_intern("symlink?"), 1, self))){
      v_path = rb_str_dup(self);
      rb_str_modify(v_path);

      while(1){
         v_path = rb_funcall(rb_cFile, rb_intern("join"), 2,
            rb_funcall(rb_cFile, rb_intern("dirname"), 1, v_path),
            rb_funcall(rb_cFile, rb_intern("readlink"), 1, v_path)
         );

         if(!RTEST(rb_funcall(rb_cFile, rb_intern("symlink?"), 1, v_path)))
            break;
      }
      
      v_path = rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, v_path);
      return path_clean(v_path);
   }
   else{
      v_pwd = rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1,
         rb_funcall(rb_cDir, rb_intern("pwd"), 0, 0)
      );
      return path_add(v_pwd, self);
   }
#endif
}

/*
* Pathname#children
* 
* Returns the children of the directory, files and subdirectories, as an
* array of Pathname objects. If you set +with_directory+ to +false+, then
* the returned pathnames will contain the filename only.
*
* Note that the result never contain the entries '.' and '..' in the
* the directory because they are not children.  Also note that this method
* is *not* recursive.
* 
* :no-doc:
* 
* In the spirit of laziness, I'm actually calling Dir.entries, and then
* iterating over them, rather than Dir.foreach.  This potentially consumes
* more memory.
*
* TODO: don't be lazy.
*/
static VALUE path_children(int argc, VALUE* argv, VALUE self){
   VALUE v_entries, v_file, v_with_dir;
   VALUE v_result = rb_ary_new();
   int i;

   rb_scan_args(argc, argv, "01", &v_with_dir);

   if(NIL_P(v_with_dir))
      v_with_dir = Qtrue;

   if(!rb_str_cmp(self, rb_str_new2(".")))
      v_with_dir = Qfalse;

   v_entries = rb_funcall(rb_cDir, rb_intern("entries"), 1, self);

   for(i = 0; i < RARRAY(v_entries)->len; i++){
      v_file = RARRAY(v_entries)->ptr[i];
      if( (!rb_str_cmp(v_file, rb_str_new2("."))) ||
          (!rb_str_cmp(v_file, rb_str_new2("..")))
      )
      {
         continue;
      }
      if(RTEST(v_with_dir)){
         rb_ary_push(v_result,
            rb_funcall(
               RBASIC(self)->klass,
               rb_intern("new"),
               1,
               rb_funcall(rb_cFile, rb_intern("join"), 2, self, v_file)
            )
         );
      }
      else{
         rb_ary_push(v_result, rb_funcall(RBASIC(self)->klass,
            rb_intern("new"), 1, v_file)
         );
      }
   }

   return v_result;
}

/* METHODS FOR WINDOWS ONLY */

/*
 * Pathname#unc?
 *
 * Determines if the string is a valid Universal Naming Convention (UNC)
 * for a server and share path.
 *
 * Windows only.
 */
static VALUE path_is_unc(VALUE self){
#ifdef ON_WINDOWS
   VALUE v_bool = Qfalse;
   
   if(PathIsUNC(TEXT(RSTRING(self)->ptr)))
      v_bool = Qtrue;

   return v_bool;
#else
   rb_raise(rb_eNotImpError, "not supported on this platform");
#endif
}

/*
 * Pathname#undecorate
 *
 * Removes the decoration from a path string.  For example,
 * C:\Path\File[5].txt would become C:\Path\File.txt.
 *
 * Windows only.  Non-destructive.
 */
static VALUE path_undecorate(VALUE self){
#ifdef ON_WINDOWS
   TCHAR path[MAX_PATH];

   _tcsncpy(path, TEXT(RSTRING(self)->ptr), RSTRING(self)->len + 1);

   PathUndecorate(path);
   
   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, rb_str_new2(path));
#else
   rb_raise(rb_eNotImpError, "not supported on this platform");
#endif
}

/*
 * Pathname#undecorate!
 *
 * Removes the decoration from a path string.  For example,
 * C:\Path\File[5].txt would become C:\Path\File.txt.  Modifies the receiver
 * in place.
 *
 * Windows only.
 */
static VALUE path_undecorate_bang(VALUE self){
#ifdef ON_WINDOWS
   TCHAR* path = TEXT(StringValuePtr(self));
     
   PathUndecorate(path);
   
   rb_funcall(self, rb_intern("replace"), 1, rb_str_new2(path));
#else
   rb_raise(rb_eNotImpError, "not supported on this platform");
#endif
   return self;
}

/*
 * Pathname#drive_number
 *
 * Returns the corresponding drive number for a given path, or nil if not
 * applicable.
 *
 * For example, Pathname.new("C:\\foo").drive_number would return 2.
 *
 * Windows only.
 */
static VALUE path_drive_number(VALUE self){
#ifdef ON_WINDOWS
   VALUE v_num = Qnil;
   int num;
    
   num = PathGetDriveNumber(TEXT(RSTRING(self)->ptr));

   if(num > -1)
      v_num = INT2FIX(num);

   return v_num;
#else
   rb_raise(rb_eNotImpError, "not supported on this platform");
#endif
}

/*
 * Pathname#long_path
 *
 * Returns the long path for a long path name.  For example,
 * C:\Progra~1\Ruby would return C:\Program Files\Ruby.
 *
 * Windows only.
*/
static VALUE path_long(VALUE self){
#ifdef ON_WINDOWS
   TCHAR path[MAX_PATH];
   
   if(!GetLongPathName(TEXT(StringValuePtr(self)), path, MAX_PATH))
      rb_raise(cPathnameError, "GetLongPathName() failed");
   
   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, rb_str_new2(path));
#else
   rb_raise(rb_eNotImpError, "not supported on this platform");
#endif
}

/*
 * Pathname#short_path
 *
 * Returns the short path for a long path name.  For example,
 * C:\Program Files\Ruby would return C:\Progra~1\Ruby.
 *
 * Windows only.
 */
static VALUE path_short(VALUE self){
#ifdef ON_WINDOWS
   TCHAR path[MAX_PATH];
   
   if(!GetShortPathName(TEXT(StringValuePtr(self)), path, MAX_PATH))
      rb_raise(cPathnameError, "GetShortPathName() failed");
   
   return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, rb_str_new2(path)); 
#else
   rb_raise(rb_eNotImpError, "not supported on this platform");
#endif 
}

/*
 * Pathname#parent
 *
 * Returns the parent directory.
 */
static VALUE path_parent(VALUE self){
   return path_add(self, rb_str_new2(".."));
}

/*
 * Pathname#relative_path_from(base)
 *
 * Returns a relative path from the argument to the receiver.  If +self+
 * is absolute, the argument must be absolute too.  If +self+ is relative,
 * the argument must be relative too.
 *
 * This method does not access the filesystem.  It assumes no symlinks.
 *
 * Raises an ArgumentError if it cannot find a relative path.
 */
static VALUE path_rel_path_from(VALUE self, VALUE v_base){
   VALUE v_dest_arr, v_base_arr, v_diff_arr, v_rel_path;
   VALUE v_dot = rb_str_new2(".");
   VALUE v_dot_dot = rb_str_new2("..");

   if(TYPE(v_base) != TYPE(self))
      v_base = rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, v_base);

   if(path_is_absolute(self) != path_is_absolute(v_base))
      rb_raise(rb_eArgError, "relative path between absolute and relative path");

   if(!rb_str_cmp(self, v_base))
      return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, rb_str_new2("."));

   if(!rb_str_cmp(v_base, v_dot))
      return self;
      
   /* Because of the way the Windows version handles Pathname#clean, we need
    * a little extra help here.
    */
#ifdef ON_WINDOWS
   if(!rb_str_cmp(v_base, v_dot_dot) &&
      (rb_str_cmp(self, v_dot) || rb_str_cmp(self, v_dot_dot))
   )
      rb_raise(rb_eArgError, "base directory may not contain '..'");
#endif

   v_dest_arr = path_to_a(path_clean(self));
   v_base_arr = path_to_a(path_clean(v_base));
   rb_ary_delete(v_dest_arr, rb_str_new2("."));
   rb_ary_delete(v_base_arr, rb_str_new2("."));

   v_diff_arr = rb_funcall(v_dest_arr, rb_intern("-"), 1, v_base_arr);

   while(RARRAY(v_base_arr)->len > 0 && RARRAY(v_dest_arr)->len > 0 &&    
      !rb_str_cmp(RARRAY(v_base_arr)->ptr[0], RARRAY(v_dest_arr)->ptr[0]))
   {
      rb_ary_shift(v_base_arr);
      rb_ary_shift(v_dest_arr);
   }

   if(rb_ary_includes(v_base_arr, v_dot_dot) == Qtrue)
      rb_raise(rb_eArgError, "base directory may not contain '..'");

   rb_funcall(v_base_arr, rb_intern("fill"), 1, v_dot_dot);

   v_rel_path = rb_funcall(v_base_arr, rb_intern("+"), 1, v_dest_arr);

   if(RARRAY(v_rel_path)->len == 0)
      return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, v_dot);
   else
      return rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1,
         rb_ary_join(v_rel_path, rb_str_new2(SEPS))
      );
}

/* Methods not handled by Facade directly */

/* IO METHODS */

/*
 * call-seq:
 *    Pathname#foreach(sep=$/){ |line| block }
 * 
 * Facade for IO.foreach, with the path automatically considered the first
 * argument.
 */
static VALUE path_foreach(int argc, VALUE* argv, VALUE self){
   VALUE v_args[2], v_block;
   
   rb_scan_args(argc, argv, "01&", &v_args[1], &v_block);

   v_args[0] = self;
   
   // This is necessary due to the special handling of nil
   if(NIL_P(v_args[1]) && argc == 0)
      v_args[1] = rb_default_rs;
   
   return rb_iterate(call_io_foreach, (VALUE)v_args, yield_block, v_block);
}

/*
 * call-seq:
 *    Pathname#read(length=nil, offset=nil)
 * 
 * Facade for IO.read, with the path automatically considered the first
 * argument.
 */
static VALUE path_read(int argc, VALUE* argv, VALUE self){
   VALUE v_length, v_offset;
   
   rb_scan_args(argc, argv, "02", &v_length, &v_offset);
   
   return rb_funcall(rb_cIO, rb_intern("read"), 3, self, v_length, v_offset);
}

/*
 * call-seq:
 *    Pathname#readlines(separator=$/)
 * 
 * Facade for IO.readlines, with the path automatically considered the first
 * argument.
 */
static VALUE path_readlines(int argc, VALUE* argv, VALUE self){
   VALUE v_sep;
   
   rb_scan_args(argc, argv, "01", &v_sep);
   
   // This is necessary due to the special handling of nil
   if(NIL_P(v_sep) && argc == 0)
      v_sep = rb_default_rs;

   return rb_funcall(rb_cIO, rb_intern("readlines"), 2, self, v_sep);  
}

/*
 * Facade for IO.sysopen
 */
static VALUE path_sysopen(int argc, VALUE* argv, VALUE self){
   VALUE v_mode, v_perm;
   
   rb_scan_args(argc, argv, "02", &v_mode, &v_perm);
   
   return rb_funcall(rb_cIO, rb_intern("sysopen"), 3, self, v_mode, v_perm);  
}

/* Dir methods */

/*
 * Facade for Dir.entries.
 */
static VALUE path_entries(VALUE self){
   int i;
   VALUE v_final = rb_ary_new();
   VALUE v_entries = rb_funcall(rb_cDir, rb_intern("entries"), 1, self);

   for(i = 0; i < RARRAY(v_entries)->len; i++){
      rb_ary_push(
         v_final,
         rb_funcall(RBASIC(self)->klass, rb_intern("new"), 1, RARRAY(v_entries)->ptr[i])
      );
   }

   return v_final;
}

/*
 * Facade for Dir.mkdir.
 */
static VALUE path_mkdir(VALUE self, VALUE* v_args){
   int num_args = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(rb_cDir, rb_intern("mkdir"), num_args, self, *v_args);
}

/*
 * Facade for Dir.open.  Note that 'opendir' is used instead of 'open' to
 * avoid conflict with the facade for File.open.
 */
static VALUE path_opendir(int argc, VALUE* argv, VALUE self){
   VALUE args[2], v_block;

   rb_scan_args(argc, argv, "01&", &args[1], &v_block);

   args[0] = self;
   
   if(rb_block_given_p())
      return rb_iterate(call_dir_open, (VALUE)args, yield_block, v_block);
   else
      return rb_funcall(rb_cDir, rb_intern("open"), 1, self);
}

/* File methods */

/*
 * Facade for File.open
 */
static VALUE path_open(int argc, VALUE* argv, VALUE self){
   VALUE args[2], v_block;

   rb_scan_args(argc, argv, "01&", &args[1], &v_block);

   /* Explicitly set args[1] to 'r' if nil */
   if(NIL_P(args[1]))
      args[1] = rb_str_new2("r");

   args[0] = self;

   if(rb_block_given_p())
      return rb_iterate(call_file_open, (VALUE)args, yield_block, v_block);
   else
      return rb_funcall(rb_cFile, rb_intern("open"), 1, self);
}

/*
 * Facade for File.chmod
 */
static VALUE path_chmod(VALUE self, VALUE v_mode){
   return rb_funcall(rb_cFile, rb_intern("chmod"), 2, v_mode, self);
}

/*
 * Facade for File.lchmod
 */
static VALUE path_lchmod(VALUE self, VALUE v_mode){
   return rb_funcall(rb_cFile, rb_intern("lchmod"), 2, v_mode, self);
}

/*
 * Facade for File.chown
 */
static VALUE path_chown(VALUE self, VALUE v_owner, VALUE v_group){
   return rb_funcall(rb_cFile, rb_intern("chown"), 2, v_owner, v_group, self);
}

/*
 * Facade for File.lchown
 */
static VALUE path_lchown(VALUE self, VALUE v_owner, VALUE v_group){
   return rb_funcall(rb_cFile, rb_intern("lchown"), 2, v_owner, v_group, self);
}

/*
 * Facade for File.fnmatch
 */
static VALUE path_fnmatch(int argc, VALUE* argv, VALUE self){
   VALUE v_pattern, v_flags;
   rb_scan_args(argc, argv, "11", &v_pattern, &v_flags);

   if(NIL_P(v_flags))
      v_flags = INT2FIX(0);

   return rb_funcall(rb_cFile, rb_intern("fnmatch?"), 3, v_pattern, self, v_flags);
}

/*
 * Facade for File.link
 */
static VALUE path_link(VALUE self, VALUE v_old){
   return rb_funcall(rb_cFile, rb_intern("link"), 2, v_old, self);
}

/*
 * Facade for File.rename
 */
static VALUE path_rename(VALUE self, VALUE v_name){
   return rb_funcall(rb_cFile, rb_intern("rename"), 2, self, v_name);
}

/*
 * Facade for File.symlink
*/
static VALUE path_symlink(VALUE self, VALUE v_old){
   return rb_funcall(rb_cFile, rb_intern("rename"), 2, v_old, self);
}

/*
 * Facade for File.truncate
 */
static VALUE path_truncate(VALUE self, VALUE v_length){
   return rb_funcall(rb_cFile, rb_intern("truncate"), 2, self, v_length);
}

/*
 * Facade for File.utime
 */
static VALUE path_utime(VALUE self, VALUE v_atime, VALUE v_mtime){
   return rb_funcall(rb_cFile, rb_intern("utime"), 3, v_atime, v_mtime, self);
}

/*
 * Facade for File.basename
 */
static VALUE path_basename(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(rb_cFile, rb_intern("basename"), n, self, *v_args);
}

/*
 * Facade for File.expand_path
 */
static VALUE path_expandpath(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(rb_cFile, rb_intern("expand_path"), n, self, *v_args);
}

/*
 * Facade for File.copy
 */
static VALUE path_copy(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(rb_cFile, rb_intern("copy"), n, self, *v_args);
}

/* FileUtils */

/*
 * Facade for FileUtils.cd
 */
static VALUE path_cd(int argc, VALUE* argv, VALUE self){
   VALUE v_options, v_block;

   rb_scan_args(argc, argv, "01&", &v_options, &v_block);

   if(NIL_P(v_options))
      v_options = rb_hash_new();

   return rb_funcall(cFileUtils, rb_intern("cd"), 1, &v_options, &v_block);
}

/*
 * Facade for FileUtils.mkdir_p
 */
static VALUE path_mkdir_p(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("mkdir_p"), n, self, *v_args);
}

/*
 * Facade for FileUtils.ln
 */
static VALUE path_ln(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("ln"), n, self, *v_args);
}

/*
 * Facade for FileUtils.ln_s
 */
static VALUE path_ln_s(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("ln_s"), n, self, *v_args);
}

/*
 * Facade for FileUtils.ln_sf
 */
static VALUE path_ln_sf(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("ln_sf"), n, self, *v_args);
}

/*
 * Facade for FileUtils.cp
 */
static VALUE path_cp(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("cp"), n, self, *v_args);
}

/*
 * Facade for FileUtils.cp_r
 */
static VALUE path_cp_r(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("cp_r"), n, self, *v_args);
}

/*
 * Facade for FileUtils.mv
 */
static VALUE path_mv(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("mv"), n, self, *v_args);
}

/*
 * Facade for FileUtils.rm
 */
static VALUE path_rm(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("rm"), n, self, *v_args);
}

/*
 * Facade for FileUtils.rm_f
 */
static VALUE path_rm_f(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("rm_f"), n, self, *v_args);
}

/*
 * Facade for FileUtils.rm_r
 */
static VALUE path_rm_r(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("rm_r"), n, self, *v_args);
}

/*
 * Facade for FileUtils.rm_rf
 */
static VALUE path_rm_rf(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("rm_rf"), n, self, *v_args);
}

/*
 * Facade for FileUtils.install
 */
static VALUE path_install(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("install"), n, self, *v_args);
}

/*
 * Facade for FileUtils.touch
 */
static VALUE path_touch(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("touch"), n, self, *v_args);
}

/*
 * Facade for FileUtils.compare_file
 */
static VALUE path_compare_file(VALUE self, VALUE v_file){
   return rb_funcall(cFileUtils, rb_intern("compare_file"), 1, self, v_file);
}

/*
 * Facade for FileUtils.uptodate?
 */
static VALUE path_uptodate(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("uptodate?"), n, self, *v_args);
}

/*
 * Facade for FileUtils.copy_file
 */
static VALUE path_copy_file(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("copy_file"), n, self, *v_args);
}

/*
 * Facade for FileUtils.remove_dir
 */
static VALUE path_remove_dir(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("remove_dir"), n, self, *v_args);
}

/*
 * Facade for FileUtils.remove_file
 */
static VALUE path_remove_file(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("remove_file"), n, self, *v_args);
}

/*
 * Facade for FileUtils.copy_entry
 */
static VALUE path_copy_entry(VALUE self, VALUE* v_args){
   int n = RARRAY(v_args)->len + 1; /* args + self */
   return rb_funcall(cFileUtils, rb_intern("copy_entry"), n, self, *v_args);
}

/*
 * call-seq:
 *    pn{ '/foo/bar' } => Same as Pathname.new('/foo/bar')
 *
 * This method provides a shortcut for creating Pathname objects.
 */
static VALUE path_pn(VALUE self){
   return rb_funcall(cPathname, rb_intern("new"), 1, rb_yield(rb_block_proc()));
}

/*
 * Prerequisites:
 *
 * * facade (available on the RAA)
 * * ftools (part of the standard library)
 * * fileutils (part of the standard library)
 *
 */
void Init_pathname2(){
   rb_require("facade");
   rb_require("ftools");
   rb_require("fileutils");
   rb_require("find");

   cPathname      = rb_define_class("Pathname", rb_cString);
   cPathnameError = rb_define_class("PathnameError", rb_eStandardError);
   cFileUtils     = rb_const_get(rb_cObject, rb_intern("FileUtils"));
   
   rb_extend_object(cPathname, rb_const_get(rb_cObject, rb_intern("Facade")));
   rb_funcall(cPathname, rb_intern("facade"), 1, rb_cFile);
   rb_funcall(cPathname, rb_intern("facade"), 1, rb_cDir);

   /* Instance methods */
   rb_define_method(cPathname, "initialize", path_initialize, 1);
   rb_define_method(cPathname, "to_a", path_to_a, 0);
   rb_define_method(cPathname, "root", path_root, 0);
   rb_define_method(cPathname, "pstrip", path_pstrip, 0);
   rb_define_method(cPathname, "pstrip!", path_pstrip_bang, 0);
   rb_define_method(cPathname, "+", path_add, 1);
   rb_define_method(cPathname, "clean", path_clean, 0);
   rb_define_method(cPathname, "clean!", path_clean_bang, 0);
   rb_define_method(cPathname, "each", path_each, 0);
   rb_define_method(cPathname, "descend", path_descend, 0);
   rb_define_method(cPathname, "ascend", path_ascend, 0);
   rb_define_method(cPathname, "children", path_children, -1);
   rb_define_method(cPathname, "<=>", path_compare, 1);
   rb_define_method(cPathname, "parent", path_parent, 0);

   rb_define_method(cPathname, "root?", path_is_root, 0);
   rb_define_method(cPathname, "relative?", path_is_relative, 0);
   rb_define_method(cPathname, "absolute?", path_is_absolute, 0);
   rb_define_method(cPathname, "relative_path_from", path_rel_path_from, 1);
   rb_define_method(cPathname, "realpath", path_realpath, 0);
   
   /* Methods for Windows only */
   rb_define_method(cPathname, "short_path", path_short, 0);
   rb_define_method(cPathname, "long_path", path_long, 0);
   rb_define_method(cPathname, "undecorate", path_undecorate, 0);
   rb_define_method(cPathname, "undecorate!", path_undecorate_bang, 0);
   rb_define_method(cPathname, "unc?", path_is_unc, 0);
   rb_define_method(cPathname, "drive_number", path_drive_number, 0);

   /* Kernel methods */
   rb_define_method(rb_mKernel, "pn", path_pn, 0);

   /* Facade methods not handled directly by Facade module */

   /* IO */
   rb_define_method(cPathname, "foreach", path_foreach, -1);
   rb_define_method(cPathname, "read", path_read, -1);
   rb_define_method(cPathname, "readlines", path_readlines, -1);
   rb_define_method(cPathname, "sysopen", path_sysopen, -1);

   /* Dir */   
   rb_define_method(cPathname, "entries", path_entries, 0);
   rb_define_method(cPathname, "mkdir", path_mkdir, -2);
   rb_define_method(cPathname, "opendir", path_opendir, -1);

   /* File */
   rb_define_method(cPathname, "open", path_open, -1);
   rb_define_method(cPathname, "chmod", path_chmod, 1);
   rb_define_method(cPathname, "lchmod", path_lchmod, 1);
   rb_define_method(cPathname, "chown", path_chown, 2);
   rb_define_method(cPathname, "lchown", path_lchown, 2);
   rb_define_method(cPathname, "fnmatch", path_fnmatch, -1);
   rb_define_method(cPathname, "fnmatch?", path_fnmatch, -1);
   rb_define_method(cPathname, "link", path_link, 1);
   rb_define_method(cPathname, "rename", path_rename, 1);
   rb_define_method(cPathname, "symlink", path_symlink, 1);
   rb_define_method(cPathname, "truncate", path_truncate, 1);
   rb_define_method(cPathname, "utime", path_utime, 2);
   rb_define_method(cPathname, "basename", path_basename, -2);
   rb_define_method(cPathname, "expand_path", path_expandpath, -2);

   /* ftools */
   rb_define_method(cPathname, "copy", path_copy, -2);

   /* FileUtils */
   rb_define_method(cPathname, "cd", path_cd, -1);
   rb_define_method(cPathname, "mkdir_p", path_mkdir_p, -2);
   rb_define_method(cPathname, "ln", path_ln, -2);
   rb_define_method(cPathname, "ln_s", path_ln_s, -2);
   rb_define_method(cPathname, "ln_sf", path_ln_sf, -2);
   rb_define_method(cPathname, "cp", path_cp, -2);
   rb_define_method(cPathname, "cp_r", path_cp_r, -2);
   rb_define_method(cPathname, "mv", path_mv, -2);
   rb_define_method(cPathname, "rm", path_rm, -2);
   rb_define_method(cPathname, "rm_f", path_rm_f, -2);
   rb_define_method(cPathname, "rm_r", path_rm_r, -2);
   rb_define_method(cPathname, "rm_rf", path_rm_rf, -2);
   rb_define_method(cPathname, "install", path_install, -2);
   rb_define_method(cPathname, "touch", path_touch, -2);
   rb_define_method(cPathname, "compare_file", path_compare_file, 1);
   rb_define_method(cPathname, "uptodate?", path_uptodate, -2);
   rb_define_method(cPathname, "copy_file", path_copy_file, -2);
   rb_define_method(cPathname, "remove_dir", path_remove_dir, -2);
   rb_define_method(cPathname, "remove_file", path_remove_file, -2);
   rb_define_method(cPathname, "copy_entry", path_copy_entry, -2);

   /* Aliases */
   rb_define_alias(cPathname, "cleanpath", "clean");
   rb_define_alias(cPathname, "rmtree", "rm_rf");
   rb_define_alias(cPathname, "mkpath", "mkdir_p");
   rb_define_alias(cPathname, "remove", "rm");
   rb_define_alias(cPathname, "rmtree", "rm_rf");
   rb_define_alias(cPathname, "/", "+");

   /* Constants */
   rb_define_const(cPathname, "VERSION", rb_str_new2("1.5.2"));
}
